@@ -1,5 +1,5 @@
 /*-
- * Copyright (C)2020 @BABOLO http://www.babolo.ru/
+ * Copyright (C)2020..2025 @BABOLO http://www.babolo.ru/
  * Inspired by su2-1.3.
  * This version written from scratch and FreeBSD's /usr/bin/su
 
@@ -60,8 +60,8 @@
  * SUCH DAMAGE.
  */
 
-#ident "@(#) Copyright (C)2020 @BABOLO http://www.babolo.ru/"
-#ident "@(#) $Id: prep.c,v 1.13 2020/10/10 18:09:01 babolo Exp $"
+#ident "@(#) Copyright (C)2020..2025 @BABOLO http://www.babolo.ru/"
+#ident "@(#) $Id: prep.c,v 1.26 2025/03/15 16:31:42 babolo Exp $"
 
 #define BLIN_COMPAT  3
 #define Bpars_COMPAT 4
@@ -69,6 +69,11 @@
 
 #include <sys/resource.h>
 #include <sys/types.h>
+
+#ifndef BLIN_OSLinux
+# include <login_cap.h>
+#endif
+
 #include <sys/event.h>
 #include <sys/stat.h>
 #include <sysexits.h>
@@ -79,10 +84,28 @@
 #include <errno.h>
 #include <paths.h>
 #include <pwd.h>
+#include <grp.h>
 #include <err.h>
+
+#ifndef BLIN_OSLinux
+# include <security/pam_appl.h>
+#endif
+
 #include <babolo/BLINflag.h>
+#include <babolo/parser.h>
 #include <mife.h>
 #include "su2.h"
+#include "su2_env.h"
+
+extern char      **environ;
+static char       *cleanenv = NULL;
+static char const *su_env[] =
+{ "MM_CHARSET", "WINDOWPATH", "ENV"  , "PWD"
+, "XAUTHORITY", "INFOPATH"  , "USER" , "HOME"
+, "LOGNAME"   , "MANPATH"   , "LANG" , "PATH"
+, "TMPDIR"    , "LC_ALL"    , "SHELL", "GROUP"
+, "SHLVL"     ,         NULL
+};
 
 static su2_pass *
 /*****************************************************************************************************
@@ -93,9 +116,9 @@ su2_copw(su2_cf *cf, su2_pass *pw) {
  ****                                                                                             ****
  *****************************************************************************************************/
 #   define blin_internal_flags (cf->flags)
-    struct passwd  *tpw = NULL;
-    size_t          stp = 0;
-    char           *p;
+    su2_pass  *tpw = NULL;
+    size_t     stp = 0;
+    char      *p;
 
     ifBLIN_QX3("+");
     if  (!pw) {
@@ -104,16 +127,18 @@ su2_copw(su2_cf *cf, su2_pass *pw) {
     }
     if  (!!pw->pw_name)   stp += strlen(pw->pw_name)   + 1;
     if  (!!pw->pw_passwd) stp += strlen(pw->pw_passwd) + 1;
+#ifndef BLIN_OSLinux
     if  (!!pw->pw_class)  stp += strlen(pw->pw_class)  + 1;
+#endif
     if  (!!pw->pw_gecos)  stp += strlen(pw->pw_gecos)  + 1;
     if  (!!pw->pw_dir)    stp += strlen(pw->pw_dir)    + 1;
     if  (!!pw->pw_shell)  stp += strlen(pw->pw_shell)  + 1;
-    if  (!(tpw = malloc(sizeof(struct passwd) + stp))) {
+    if  (!(tpw = malloc(sizeof(su2_pass) + stp))) {
         ifBLIN_QW0("Malloc failed #3");
         goto out;
     }
-    bcopy(pw, tpw, sizeof(struct passwd));
-    p = (char*)(void*)tpw + sizeof(struct passwd);
+    bcopy(pw, tpw, sizeof(su2_pass));
+    p = (char*)(void*)tpw + sizeof(su2_pass);
     if  (!!pw->pw_name) {
         tpw->pw_name = p;
         p = stpcpy(p, pw->pw_name) + 1;
@@ -122,10 +147,12 @@ su2_copw(su2_cf *cf, su2_pass *pw) {
         tpw->pw_passwd = p;
         p = stpcpy(p, pw->pw_passwd) + 1;
     }
+#ifndef BLIN_OSLinux
     if  (!!pw->pw_class) {
         tpw->pw_class = p;
         p = stpcpy(p, pw->pw_class) + 1;
     }
+#endif
     if  (!!pw->pw_gecos) {
         tpw->pw_gecos = p;
         p = stpcpy(p, pw->pw_gecos) + 1;
@@ -157,9 +184,9 @@ su2_iniud(su2_cf *cf, char c, const char *uname, su2_pass **pw, uid_t uid) {
  *****************************************************************************************************
  *****************************************************************************************************/
 #   define blin_internal_flags (cf->flags)
-    int              ex = EX_OK;
-    char            *tail;
-    struct passwd   *tpw;
+    int         ex = EX_OK;
+    char       *tail;
+    su2_pass   *tpw;
 
     ifBLIN_QX3("+ -%c %s %d", c, !uname ? "" : uname, uid);
     *pw = NULL;
@@ -180,7 +207,7 @@ su2_iniud(su2_cf *cf, char c, const char *uname, su2_pass **pw, uid_t uid) {
             goto ready;
         }
         errno = 0;
-        uid = strtoul(uname, &tail, 0);
+        uid = (uid_t)strtoul(uname, &tail, 0);
         if  (errno) {
             ifBLIN_QW0("-%c strtoul=%s~", c, uname);
             ex = EX_USAGE;
@@ -204,7 +231,7 @@ next:
         goto out;
     }
 ready:
-    if  (!(*pw = malloc(sizeof(struct passwd)))) {
+    if  (!(*pw = malloc(sizeof(su2_pass)))) {
         ifBLIN_QW0("Malloc failed #2");
         ex = EX_OSERR;
         goto out;
@@ -293,7 +320,7 @@ su2_dump(su2_cf *cf) {
               , !!cf->command ? cf->command : ""
               , !!cf->command ? '~' : 0
               );
-    for (int i = 0; (i < cf->amax) || !!cf->args[i]; ++i) {
+    for (uint i = 0; (i < cf->amax) || !!cf->args[i]; ++i) {
         ifBLIN_QX0( "%c[%3u]%s%s~"
                   , (i == cf->argp) ? ((i == cf->amax) ? '#' : '*') : ((i == cf->amax) ? '+' : ' ')
                   , i
@@ -323,7 +350,7 @@ su2_tty(su2_cf *cf) {
         if  (!!isatty(i) && !!(o = ttyname(i))) {
             cf->ttyfd = i;
             if  (!(cf->tty = strdup(o))) {
-                ifBLIN_QW0("strdup", o);
+                ifBLIN_QW0("strdup=%s~", o);
                 ex = EX_OSERR;
                 goto out;
             }
@@ -351,41 +378,44 @@ su2_shell(su2_cf *cf) {
     const char   *o;
 
     ifBLIN_QX3("+");
-    setusershell();
-    if  (!!cf->topw->pw_uid && !!cf->mypw->pw_uid) {
-        for (;;) {
-            if  (!(o = getusershell())) {
-                cf->shell = _PATH_BSHELL;
-                break;
-            }
-            if  (!strcmp((SU2_LOGIN & cf->flags) ? cf->topw->pw_shell : cf->mypw->pw_shell, o)) {
-                cf->shell = (SU2_LOGIN & cf->flags) ? cf->topw->pw_shell : cf->mypw->pw_shell;
-                break;
-        }   }
-        endusershell();
-    } else {
-        cf->shell = (SU2_LOGIN & cf->flags) ? cf->topw->pw_shell : cf->mypw->pw_shell;
-    }
-    o = strrchr(cf->shell, '/');
-    if  (o) ++o; else o = cf->shell;
-    if  ((SU2_DASH & cf->flags)) {
-        /* NOTHING */
-    } else {
-        if  (!!cf->command) {
-            cf->args[cf->argp] = cf->command;
-            cf->amax++;
-            cf->args[--cf->argp] = "-c";
-        }
-        o = strrchr(cf->shell, '/');
-        if  (o) ++o; else o = cf->shell;
-        if  (SU2_LOGIN & cf->flags) {
-            cf->args[--cf->argp] = "-su2";
-        } else if (strcmp(o, "csh") || strcmp(o, "tcsh")) {
-            cf->args[--cf->argp] = "-m";
-            cf->args[--cf->argp] = "_su2";
+    if  (!cf->xfile) {
+        setusershell();
+        if  (!!cf->topw->pw_uid && !!cf->mypw->pw_uid) {
+            for (;;) {
+                if  (!(o = getusershell())) {
+                    cf->shell = _PATH_BSHELL;
+                    break;
+                }
+                if  (!strcmp((SU2_LOGIN & cf->flags) ? cf->topw->pw_shell : cf->mypw->pw_shell, o)) {
+                    cf->shell = (SU2_LOGIN & cf->flags) ? cf->topw->pw_shell : cf->mypw->pw_shell;
+                    break;
+            }   }
+            endusershell();
         } else {
-            cf->args[--cf->argp] = "su2";
-    }   }
+            cf->shell = (SU2_LOGIN & cf->flags) ? cf->topw->pw_shell : cf->mypw->pw_shell;
+        }
+        if  (!(SU2_DASH & cf->flags)) {
+            if  (!!cf->command) {
+                cf->args[cf->argp] = cf->command;
+                cf->amax++;
+                cf->args[--cf->argp] = "-c";
+            }
+            o = strrchr(cf->shell, '/');
+            if  (o) ++o; else o = cf->shell;
+            if  (SU2_LOGIN & cf->flags) {
+                cf->args[--cf->argp] = "-su2";
+            } else if (strcmp(o, "csh") || strcmp(o, "tcsh")) {
+                cf->args[--cf->argp] = "-m";
+                cf->args[--cf->argp] = "_su2";
+            } else {
+                cf->args[--cf->argp] = "su2";
+        }   }
+    } else {
+        cf->args[cf->argp] = NULL;
+        cf->args[--cf->argp] = cf->xfile;
+        cf->args[--cf->argp] = "/bin/sh";
+        cf->shell = _PATH_BSHELL;
+    }
     ifBLIN_QX3("- %d", ex);
     return(ex);
 #   undef blin_internal_flags
@@ -396,30 +426,29 @@ static su2_found
  ****                                                ****
  *****************************************************************************************************
  ****                                                                                             ****/
-su2_finduser(su2_cf *cf, char *path) {                                                           /****
+su2_finduser(su2_cf *cf, char const *path) {                                                     /****
  ****                                                                                             ****
  *****************************************************************************************************/
 #   define blin_internal_flags (cf->flags)
     su2_found        ex = SU2_FNDNONE;
     char            *window;
-    mife_descriptor *md;
     char            *p;
 
     ifBLIN_QX3("+ %s %s", path, cf->frompw->pw_name);
-    if  (!(md = mife_init(MIFE_FULL | MIFE_BUFX | 0x0F01))) {
+    if  (!(cf->md = mife_init(MIFE_FULL | MIFE_BUFX | 0x0F01))) {
         ifBLIN_QW0("mife_init");
         ex = SU2_FNDFAIL;
         goto out;
     }
-    if  (0 > mife_ctlfile(md, path)) {
+    if  (0 > mife_ctlfile(cf->md, path)) {
         ifBLIN_QW0("mife_ctlfile %s", path);
         ex = SU2_FNDFAIL;
-        goto close;
+        goto out;
     }
-    if  (!(window = mife_window(md, 0, 0))) {
+    if  (!(window = mife_window(cf->md, 0, 0))) {
         ifBLIN_QW0("mife_window %s", path);
         ex = SU2_FNDFAIL;
-        goto close;
+        goto out;
     }
     for (p = strstr(window, "+*"); !!p; p = strstr(p, "+*")) {
         ifBLIN_QX4("full %"BLIN_D, BLIN_I(p - window));
@@ -427,7 +456,7 @@ su2_finduser(su2_cf *cf, char *path) {
         if  ((p == window) || (' ' >= *(p - 1))) {
             ifBLIN_QX4("FULL %"BLIN_D, BLIN_I(p - window));
             ex = SU2_FNDFULL;
-            goto close;
+            goto out;
     }   }
     for (p = strchr(window, '*'); !!p; p = strchr(p, '*')) {
         ifBLIN_QX4("star %"BLIN_D, BLIN_I(p - window));
@@ -435,7 +464,7 @@ su2_finduser(su2_cf *cf, char *path) {
         if  ((p == window) || (' ' >= *(p - 1))) {
             ifBLIN_QX4("STAR %"BLIN_D, BLIN_I(p - window));
             ex = SU2_FNDSTAR;
-            goto close;
+            goto out;
     }   }
     for (p = strstr(window, cf->frompw->pw_name); !!p; p = strstr(p, cf->frompw->pw_name)) {
         ifBLIN_QX4("user %"BLIN_D, BLIN_I(p - window));
@@ -443,20 +472,18 @@ su2_finduser(su2_cf *cf, char *path) {
         if  (p == window) {
             ifBLIN_QX4("USER %"BLIN_D, BLIN_I(p - window));
             ex = SU2_FNDUSER;
-            goto close;
+            goto out;
         }
         if  ('+' == *(p - 1)) {
             ifBLIN_QX4("PLUS %"BLIN_D, BLIN_I(p - window));
             ex = SU2_FNDPLUS;
-            goto close;
+            goto out;
         }
         if  (' ' >= *(p - 1)) {
             ifBLIN_QX4("USER %"BLIN_D, BLIN_I(p - window));
             ex = SU2_FNDUSER;
-            goto close;
+            goto out;
     }   }
-close:
-    if  (!!mife_fini(md)) ifBLIN_QW0("mife_fini %s", path);
 out:
     ifBLIN_QX3("- %02X", ex);
     return(ex);
@@ -517,3 +544,130 @@ out:
     return(ex);
 #   undef blin_internal_flags
 }
+
+int
+/*****************************************************************************************************
+ *****************************************************************************************************
+ ****     fork()                                                           ****
+ *****************************************************************************************************
+ ****                                                                                             ****/
+su2_posenv( su2_cf *cf                                                                           /****/
+#ifndef BLIN_OSLinux
+          , login_cap_t *lc                                                                      /****/
+#endif
+          ) {                                                                                    /****
+ ****                                                                                             ****
+ *****************************************************************************************************
+ *****************************************************************************************************/
+#   define blin_internal_flags (cf->flags)
+    char         **envpam;
+    struct group  *grent;
+    int            ex = EX_OK;
+    const char    *p;
+    char         **e;
+    char          *v;
+
+    ifBLIN_QX3("+ %"BLIN_X, BLIN_I(cf));
+#ifdef BLIN_OSLinux
+    if  (!!setgid(cf->topw->pw_gid)) {
+        ifBLIN_QW0("setgid %u", cf->topw->pw_gid);
+        ex = EX_NOPERM;
+        goto out;
+    }
+    if  (!!setuid(cf->topw->pw_uid)) {
+        ifBLIN_QW0("setuid %u", cf->topw->pw_uid);
+        ex = EX_NOPERM;
+        goto out;
+    }
+#else
+    if  (0 > setusercontext( lc
+                           , cf->topw
+                           , cf->topw->pw_uid
+                           , LOGIN_SETUSER | LOGIN_SETCPUMASK | LOGIN_SETLOGINCLASS
+        )                  ) {
+        ifBLIN_QW0("setusercontext(SETUSER SETCPUMASK SETLOGINCLASS)");
+        ex = EX_OSERR;
+        goto out;
+    }
+#endif
+    ifBLIN_QX6("uid %d", getuid());
+    if  (SU2_LOGIN & cf->flags) {
+        p = getenv("TERM");
+        environ = &cleanenv;
+        if  (!!setenv("HOME", cf->topw->pw_dir, 1)) {
+            ifBLIN_QW1("setenv(HOME, %s, 1)", cf->topw->pw_dir);
+        }
+        if  (!!setenv("SHELL", cf->shell, 1)) ifBLIN_QW1("setenv(SHELL, %s, 1)", cf->shell);
+#ifndef BLIN_OSLinux
+        if  (!!(envpam = pam_getenvlist(cf->pamh))) {
+            for (e = envpam; !!*e; ++e) {
+                if  (1024 < strlen(*e)) continue;
+                if  (!strncmp(*e, "LD_", 3)) continue;
+                if  (!(v = strchr(*e, '='))) continue;
+                *(v++) = '\0';
+                if  (1 == babolo_testword(&env, *e)) continue;
+                if  (!!setenv(*e, v, 1)) ifBLIN_QW1("setenv(setenv(%s, %s, 1)", *e, v);
+        }   }
+        if  (0 > setusercontext( lc
+                               , cf->topw
+                               , cf->topw->pw_uid
+                               , LOGIN_SETPATH | LOGIN_SETUMASK | LOGIN_SETENV | LOGIN_SETRESOURCES
+            )                  ) {
+            ifBLIN_QW0("setusercontext(LOGIN_SETPATH LOGIN_SETUMASK LOGIN_SETENV)");
+            ex = EX_OSERR;
+            goto out;
+        }
+#endif
+        if  (!!p && !!setenv("TERM", p, 1)) ifBLIN_QW1("setenv(TERM, %s, 1)", p);
+        p = NULL;
+#ifndef BLIN_OSLinux
+        p = pam_getenv(cf->pamh, "HOME");
+#endif
+        if  (!!chdir(p ? p : cf->topw->pw_dir)) {
+            ifBLIN_QW0("chdir %s", p ? p : cf->topw->pw_dir);
+            ex = EX_UNAVAILABLE;
+            goto out;
+    }   }
+    for (int i = 0; !!su_env[i]; ++i) {
+        if  (!(p = getenv(su_env[i]))) {
+            ifBLIN_QW1("getenv %s", su_env[i]);
+        } else {
+            asprintf(&v, "SU2_%s", su_env[i]);
+            if  (!v) {
+                ifBLIN_QW1("asprintf %s", su_env[i]);
+            } else {
+                if  (!!setenv(v, p, 1)) ifBLIN_QW1("setenv(%s, %s, 1)", v, p);
+                free(v);
+    }   }   }
+    asprintf(&v, "%u", cf->frompw->pw_uid);
+    if  (!v) {
+        ifBLIN_QW1("asprintf frompw->pw_uid");
+    } else {
+        if  (!!setenv("SU2_REAL_UID", v, 1)) ifBLIN_QW1("setenv(SU2_REAL_UID, %s, 1)", v);
+        free(v);
+    }
+    asprintf(&v, "%u", cf->frompw->pw_gid);
+    if  (!v) {
+        ifBLIN_QW1("asprintf frompw->pw_gid");
+    } else {
+        if  (!!setenv("SU2_REAL_GID", v, 1)) ifBLIN_QW1("setenv(SU2_REAL_GID, %s, 1)", v);
+        free(v);
+    }
+    if  (!!setenv("SU2_REAL_UNAME", cf->frompw->pw_name, 1)) {
+        ifBLIN_QW1("setenv(SU2_REAL_UNAME, %s, 1)", cf->frompw->pw_name);
+    }
+    errno = 0;
+    if  (!(grent = getgrgid(cf->frompw->pw_gid))) {
+        if  (!errno) {
+            ifBLIN_QX1("No group %u", cf->frompw->pw_gid);
+        } else {
+            ifBLIN_QW1("getgrgid %u", cf->frompw->pw_gid);
+        }
+    } else if (!!setenv("SU2_REAL_GNAME", grent->gr_name, 1)) {
+        ifBLIN_QW1("setenv(SU2_REAL_GNAME, %s, 1)", grent->gr_name);
+    }
+out:
+    ifBLIN_QX3("- %d", ex);
+    return(ex);
+#   undef blin_internal_flags
+}
