@@ -1,5 +1,5 @@
 /*-
- * Copyright (C)2020 @BABOLO http://www.babolo.ru/
+ * Copyright (C)2020..2025 @BABOLO http://www.babolo.ru/
  * Inspired by su2-1.3.
  * This version written from scratch and FreeBSD's /usr/bin/su
  *
@@ -60,16 +60,32 @@
  * SUCH DAMAGE.
  */
 
-#ident "@(#) Copyright (C)2020 @BABOLO http://www.babolo.ru/"
-#ident "@(#) $Id: su2.c,v 1.19 2020/10/11 14:23:30 babolo Exp $"
+#ident "@(#) Copyright (C)2020..2025 @BABOLO http://www.babolo.ru/"
+#ident "@(#) $Id: su2.c,v 1.43 2025/12/21 19:16:40 babolo Exp $"
 
 #define BLIN_COMPAT  3
 #define Bpars_COMPAT 4
 #define MIFE_COMPAT  5
 
-#include <sys/param.h>
-#include <login_cap.h>
+#include <sys/mount.h>
+#include <sys/types.h>
+#include <sys/event.h>
+
+#ifdef BLIN_OSLinux
+# include <sys/vfs.h>
+# include <sys/statvfs.h>
+#else
+# include <login_cap.h>
+#endif
+
+#include <sys/stat.h>
 #include <sys/wait.h>
+#include <sys/uio.h>
+
+#ifndef BLIN_OSLinux
+# include <sys/ktrace.h>
+#endif
+
 #include <sysexits.h>
 #include <string.h>
 #include <stdlib.h>
@@ -78,32 +94,38 @@
 #include <signal.h>
 #include <stdio.h>
 #include <errno.h>
+#include <fcntl.h>
 #include <pwd.h>
 #include <err.h>
-#include <security/pam_appl.h>
+
+#ifndef BLIN_OSLinux
+# include <security/pam_appl.h>
+#endif
+
 #include <babolo/BLINflag.h>
 #include <babolo/parser.h>
+#include <mife.h>
 #include "su2.h"
 #include "su2_env.h"
 
-extern char **environ;
-static char  *cleanenv = NULL;
-
-static void
+__attribute__((__noreturn__)) static void
 usage(int ex) {
     fprintf( stderr
            , "su2 @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
              "     Switch user rights\n"
-             "Usage: su2 [-cf] [-u fromUID] [-nice] [toUID] [-c command | -- prog args...]\n"
+             "Usage: su2 [-OPTIONS] [toUNAME] [-c command | -- prog args...]\n"
              "    -- ...      - execve(...)\n"
              "    -           - login\n"
-             "    -c command  - exec 'sh -c command'\n"
+             "    -a UNAME    - use UNAMEs(or UIDs) access rights\n"
+             "    -c COMMAND  - exec '/bin/sh -c command'\n"
              "    -d          - run /bin/csh\n"
-             "    -h          - \n"
-             "    -n nice     - nice\n"
+             "    -f FILE     - run /bin/sh file\n"
+             "    -l          - login\n"
+             "    -n NICE     - nice\n"
              "    -q          - quiet\n"
              "    -s          - register new UID in utmp\n"
-             "    -u fromUID  - whom rights used before user switsh\n"
+             "    -t CLASS    - login class\n"
+             "    -u UNAME    - switsh to UNAME(or UID), mandatory password with -f\n"
              "    -v[v[v...]] - verbose\n"
            )
     ;
@@ -116,21 +138,47 @@ main(int argc, char *argv[]) {
     struct sigaction     sa_pipe;
     struct sigaction     sa_quit;
     struct sigaction     sa_int;
-    char               **envpam;
     pid_t                gchild;
     pid_t                child;
-    const char          *flar    = "-a:c:dln:qst:u:v=:?";
+    const char          *flar    = "-a:c:df:ln:qst:u:v=:?";
     int                  temp;
     int                  fds[2];
+    int                  xfd;
     pid_t                pid;
     babolo_opts         *bos     = NULL;
     su2_cf              *cf      = NULL;
     struct sigaction     sa;
+#ifndef BLIN_OSLinux
     login_cap_t         *lc;
+#endif
     int                  ex      = EX_OK;
     const char          *o;
     int                  c;
 
+#if 0
+    {   int f = open("ktrace.in", O_RDWR | O_TRUNC | O_CREAT, 0666);
+        close(f);
+        ktrace( "ktrace.in"
+              , KTROP_SET
+              , KTRFAC_SYSCALL
+              | KTRFAC_SYSRET
+              | KTRFAC_NAMEI
+              | KTRFAC_GENIO
+              | KTRFAC_PSIG
+              | KTRFAC_CSW
+              | KTRFAC_USER
+              | KTRFAC_STRUCT
+              | KTRFAC_SYSCTL
+              | KTRFAC_PROCCTOR
+              | KTRFAC_PROCDTOR
+              | KTRFAC_CAPFAIL
+              | KTRFAC_FAULT
+              | KTRFAC_FAULTEND
+              | KTRFAC_INHERIT
+              , getpid()
+              );
+    }
+#endif
     if  (!(cf = calloc(1, sizeof(*cf)))) {
         warnx("Malloc failed #1");
         errno = ENOMEM;
@@ -140,6 +188,13 @@ main(int argc, char *argv[]) {
 #   define blin_internal_flags (cf->flags)
     openlog("su2", LOG_CONS, LOG_AUTH);
     BLIN_VERBOSE(cf->flags);
+    cf->arsz = argc + ARGSHIFT + 3;
+    if  (!(cf->args = malloc(cf->arsz * sizeof(char *)))) {
+        ifBLIN_QW0("Malloc failed #2");
+        errno = ENOMEM;
+        ex = EX_OSERR;
+        goto out;
+    }
     if  (!(bos = babolo_openopts(Bpars_NONU | Bpars_SYME, 'Z'))) {
         ifBLIN_QW0("babolo_openopts");
         ex = EX_SOFTWARE;
@@ -155,7 +210,7 @@ main(int argc, char *argv[]) {
         ifBLIN_QX4("-%c(%02X)%02X", c ? c : 1, c, bos->c);
         if  (!c) break;
         if  (0 > c) {
-            ifBLIN_QW0("babolo_getopts -%c", bos->c);
+            ifBLIN_QW0("babolo_getopts -%c(%02X)", bos->c, bos->c & 0x00FF);
             ex = EX_USAGE;
             goto out;
         }
@@ -186,14 +241,27 @@ main(int argc, char *argv[]) {
                 ;     ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cf->command, o);
                 ; }
                 ; if  (!(cf->command = strdup(o))) {
-                ;     ifBLIN_QW0("strdup", o);
+                ;     ifBLIN_QW0("strdup=%s~", o);
                 ;     ex = EX_OSERR;
                 ;     goto out;
                 ; }
                 ; break;
         case 'd': cf->flags |= SU2_CSH;
                 ; break;
-        case 'n': {   int n;
+        case 'f': if  (!(o = babolo_getoptsarg(bos))) {
+                ;     ifBLIN_QW0("No -%c arg", bos->c);
+                ;     ex = EX_USAGE;
+                ;     goto out;
+                ; }
+                ; if  (!!cf->xfile) ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cf->xfile, o);
+                ; if  (!(cf->xfile = strdup(o))) {
+                ;     ifBLIN_QW0("strdup=%s~", o);
+                ;     ex = EX_OSERR;
+                ;     goto out;
+                ; }
+                ; cf->flags |= SU2_XFILE;
+                ; break;
+        case 'n': {   int n = 0;
 
                 ;     if  (SU2_NICE & cf->flags) {
                 ;         cf->flags |= SU2_RENICE;
@@ -201,7 +269,7 @@ main(int argc, char *argv[]) {
                 ;     }
                 ;     cf->flags |= SU2_NICE;
                 ;     errno = 0;
-                ;     cf->nice = babolo_getoptunum(bos, 0);
+                ;     cf->nice = (int)babolo_getoptsnum(bos, 0);
                 ;     if  (!!errno) {
                 ;         ifBLIN_QW0("babolo_getoptunum");
                 ;         ex = EX_USAGE;
@@ -234,7 +302,7 @@ main(int argc, char *argv[]) {
                 ;     ifBLIN_QX1("Redefined -%c option: %s->%s", bos->c, cf->topw->pw_name, o);
                 ; }
                 ; if  (!!su2_iniud(cf, bos->c, o, &cf->topw, 0)) {
-                ;     ifBLIN_QW0("No %s fromUID", o);
+                ;     ifBLIN_QW0("No %s toUID", o);
                 ;     ex = EX_USAGE;
                 ;     goto out;
                 ; }
@@ -262,14 +330,24 @@ main(int argc, char *argv[]) {
     ifBLIN_QX4("%08X %08X", bos->flags, Bpars_DASH);
     if  (Bpars_DASH & bos->flags) cf->flags |= SU2_DASH;
     cf->argp = cf->amax = ARGSHIFT;
-    for (int i = cf->argp; i < MAXARGS;) {
-        if  (!(o = babolo_getargs(bos))) break;
-        if  (!(cf->args[i] = strdup(o))) {
+    for (cf->args[cf->amax] = NULL; ;) {
+        if  (  (cf->arsz <= cf->amax)
+            && !(cf->args = realloc(cf->args, (cf->arsz *= 2) * sizeof(char *)))
+            ) {
+            ifBLIN_QW0("Realloc failed");
+            ex = EX_OSERR;
+            goto out;
+        }
+        o = babolo_getargs(bos);
+        if  (!o) {
+            cf->args[cf->amax++] = NULL;
+            break;
+        } else if (!(cf->args[cf->amax++] = strdup(o))) {
             ifBLIN_QW0("strdup");
             ex = EX_OSERR;
             goto out;
         }
-        cf->amax = ++i;
+        ifBLIN_QX4("[%d]=%s~", cf->amax - 1, cf->args[cf->amax - 1]);
     }
     if  (!!geteuid()) {
         ifBLIN_QX0("Not running setuid root");
@@ -288,62 +366,135 @@ main(int argc, char *argv[]) {
         ex = EX_OSERR;
         goto out;
     }
-    if  (!(SU2_DASH & cf->flags)) {
+    ifBLIN_QX2("%08X", cf->flags);
+    if  (SU2_XFILE & cf->flags) { /*  su2 -f, ,  */
+        struct stat   sb;
+        struct statfs mb;
+
+    ;   if  (0 > (xfd = open(cf->xfile, O_RDONLY
+#ifndef BLIN_OSLinux
+                                      | O_SHLOCK
+#endif
+            )    )          ) {
+    ;       ifBLIN_QW0("Script %s open", cf->xfile);
+    ;       ex = EX_IOERR;
+    ;       goto out;
+    ;   }
+    ;   if  (!!fstat(xfd, &sb)) {
+    ;       ifBLIN_QW0("Script %s fstat", cf->xfile);
+    ;       ex = EX_IOERR;
+    ;       goto out;
+    ;   }
+    ;   if  (!!fstatfs(xfd, &mb)) {
+    ;       ifBLIN_QW0("Script %s statfs", cf->xfile);
+    ;       ex = EX_IOERR;
+    ;       goto out;
+    ;   }
+    ;   if  (!(S_ISUID & sb.st_mode) || (
+#ifdef BLIN_OSLinux
+                                         (MS_NOEXEC | ST_NOSUID)
+#else
+                                         (MNT_NOSUID | MNT_NOEXEC)
+#endif
+                                        & mb.f_flags
+            )                           ) {
+    ;       cf->flags &= ~SU2_XFILE;
+    ;       if  (!cf->topw && !!su2_iniud(cf, ':', "root", &cf->topw, 0)) {
+    ;           ifBLIN_QW0("No root toUID");
+    ;           ex = EX_NOPERM;
+    ;           goto out;
+    ;       }
+    ;   } else if ((S_IWOTH | S_IWGRP | S_IWUSR) & sb.st_mode) {
+    ;       ifBLIN_QX1("SUID excess permissions on script");
+    ;       cf->flags &= ~SU2_XFILE;
+    ;   } else if (  ((sb.st_uid != cf->mypw->pw_uid) || !(S_IXUSR & sb.st_mode))
+                  && ((sb.st_gid != cf->mypw->pw_gid) || !(S_IXGRP & sb.st_mode))
+                  && !(S_IXOTH & sb.st_mode)
+                  ) {
+    ;       ifBLIN_QX1( "SUID permission denied %u:%u %u:%u %08o %08o %08o %08o"
+                      , cf->mypw->pw_uid
+                      , cf->mypw->pw_gid
+                      , sb.st_uid
+                      , sb.st_gid
+                      , sb.st_mode
+                      , S_IXUSR
+                      , S_IXGRP
+                      , S_IXOTH
+                      );
+    ;       cf->flags &= ~SU2_XFILE;
+    ;   } else if (!!cf->topw) {
+    ;       if  (cf->topw->pw_uid != sb.st_uid) {
+    ;           ifBLIN_QX1("SUID incorrect permissions in script");
+    ;           cf->flags &= ~SU2_XFILE;
+    ;       }
+    ;   } else {
+    ;       cf->flags |= SU2_XPASS;
+    ;       if  (!!su2_iniud(cf, 0, NULL, &cf->topw, sb.st_uid)) {
+    ;           ifBLIN_QW0("No %u toUID", sb.st_uid);
+    ;           ex = EX_NOPERM;
+    ;           goto out;
+    ;   }   }
+        ifBLIN_QX2("%08X", cf->flags);
+    } else if (!(SU2_DASH & cf->flags)) {
         ifBLIN_QX4("DASH 0=%s~", !cf->args[ARGSHIFT] ? "" : cf->args[ARGSHIFT]);
-        if  (!cf->topw && !!su2_iniud(cf, '-', cf->args[ARGSHIFT], &cf->topw, 0)) {
-            ifBLIN_QW0("No %s toUID", !cf->args[ARGSHIFT] ? "root" : cf->args[ARGSHIFT]);
-            ex = EX_USAGE;
-            goto out;
-        }
-        if  (!!cf->args[ARGSHIFT]) ++cf->argp;
-        if  (!!cf->args[ARGSHIFT + 1]) {
-        ;   ifBLIN_QX4("1=%s~", cf->args[ARGSHIFT + 1]);
-        ;   if  (!strncmp(cf->args[ARGSHIFT + 1], "-c", 2)) {
-        ;       ifBLIN_QX4("-c");
-        ;       if  (2 < strnlen(cf->args[ARGSHIFT + 1], 3)) {
-        ;           if  (!!cf->command) {
-        ;               ifBLIN_QX1( "Command redefined from=%s~ to=%s~"
+    ;   if  (!cf->topw && !!su2_iniud(cf, '-', cf->args[ARGSHIFT], &cf->topw, 0)) {
+    ;       ifBLIN_QW0("No %s toUID", !cf->args[ARGSHIFT] ? "root" : cf->args[ARGSHIFT]);
+    ;       ex = EX_USAGE;
+    ;       goto out;
+    ;   }
+    ;   if  (!!cf->args[ARGSHIFT]) {
+    ;       ifBLIN_QX6("%"BLIN_X, BLIN_I(cf->args[ARGSHIFT]));
+    ;       ifBLIN_QX4("%d[%d]=%s~", cf->argp, ARGSHIFT, cf->args[ARGSHIFT]);
+    ;       ++cf->argp;
+    ;   }
+    ;   if  (!!cf->args[ARGSHIFT + 1]) {
+    ;       ifBLIN_QX6("%"BLIN_X, BLIN_I(cf->args[ARGSHIFT + 1]));
+    ;       ifBLIN_QX4("1=%s~", cf->args[ARGSHIFT + 1]);
+    ;       if  (!strncmp(cf->args[ARGSHIFT + 1], "-c", 2)) {
+    ;           ifBLIN_QX4("-c");
+    ;           if  (2 < strnlen(cf->args[ARGSHIFT + 1], 3)) {
+    ;               if  (!!cf->command) {
+    ;                   ifBLIN_QX1( "Command redefined from=%s~ to=%s~"
                                   , cf->command
                                   , &cf->args[ARGSHIFT + 1][2]
                                   );
-        ;           }
-        ;           if  (!(cf->command = strdup(&cf->args[ARGSHIFT + 1][2]))) {
-        ;               ifBLIN_QW0("strdup", &cf->args[ARGSHIFT + 1][2]);
-        ;               ex = EX_OSERR;
-        ;               goto out;
-        ;           }
-        ;       } else if (!!cf->args[ARGSHIFT + 2]) {
-        ;           ifBLIN_QX4("2=%s~", cf->args[ARGSHIFT + 2]);
-        ;           if  (!!cf->command) {
-        ;               ifBLIN_QX1( "Command redefined from=%s~ to=%s~"
+    ;               }
+    ;               if  (!(cf->command = strdup(&cf->args[ARGSHIFT + 1][2]))) {
+    ;                   ifBLIN_QW0("strdup=%s~", &cf->args[ARGSHIFT + 1][2]);
+    ;                   ex = EX_OSERR;
+    ;                   goto out;
+    ;               }
+    ;           } else if (!!cf->args[ARGSHIFT + 2]) {
+    ;               ifBLIN_QX4("2=%s~", cf->args[ARGSHIFT + 2]);
+    ;               if  (!!cf->command) {
+    ;                   ifBLIN_QX1( "Command redefined from=%s~ to=%s~"
                                   , cf->command
                                   , cf->args[ARGSHIFT + 2]
                                   );
-        ;           }
-        ;           if  (!(cf->command = strdup(cf->args[ARGSHIFT + 2]))) {
-        ;               ifBLIN_QW0("strdup", cf->args[ARGSHIFT + 2]);
-        ;               ex = EX_OSERR;
-        ;               goto out;
-        ;           }
-        ;       } else {
-        ;           ifBLIN_QX0("Command omited");
-        ;           ex = EX_USAGE;
-        ;           goto out;
-        ;       }
-        ;   } else if (!!strcmp(cf->args[ARGSHIFT + 1], "--")) {
-        ;       ifBLIN_QX0("No --");
-        ;       ex = EX_USAGE;
-        ;       goto out;
-        ;   } else {
-        ;       cf->flags |= SU2_DASH;
-        ;       ++cf->argp;
+    ;               }
+    ;               if  (!(cf->command = strdup(cf->args[ARGSHIFT + 2]))) {
+    ;                   ifBLIN_QW0("strdup=%s~", cf->args[ARGSHIFT + 2]);
+    ;                   ex = EX_OSERR;
+    ;                   goto out;
+    ;               }
+    ;           } else {
+    ;               ifBLIN_QX0("Command omited");
+    ;               ex = EX_USAGE;
+    ;               goto out;
+    ;           }
+    ;       } else if (!!strcmp(cf->args[ARGSHIFT + 1], "--")) {
+    ;           ifBLIN_QX0("No --");
+    ;           ex = EX_USAGE;
+    ;           goto out;
+    ;       } else {
+    ;           cf->flags |= SU2_DASH;
+    ;           ++cf->argp;
         }   }
     } else if (!cf->topw && !!su2_iniud(cf, ':', "root", &cf->topw, 0)) {
         ifBLIN_QW0("No root toUID");
-        ex = EX_USAGE;
+    ;   ex = EX_USAGE;
         goto out;
     }
-    endpwent();
     if  (!cf->frompw) {
         ifBLIN_QX0("Unknown fromUID");
         ex = EX_USAGE;
@@ -354,6 +505,10 @@ main(int argc, char *argv[]) {
         ex = EX_USAGE;
         goto out;
     }
+    /* cf->mypw   --                                            *
+     * cf->frompw --   ,     *
+     * cf->topw   --                                      */
+    ifBLIN_QX2("my %u, from %u, to %u", cf->mypw->pw_uid, cf->frompw->pw_uid, cf->topw->pw_uid);
     if  (!!(ex = su2_shell(cf))) {
         ifBLIN_QX0("su2_shell");
         goto out;
@@ -371,47 +526,62 @@ main(int argc, char *argv[]) {
         ifBLIN_QX0("su2_rc");
         goto out;
     }
-    if  (!((SU2_NOCHCK | SU2_UFOUND) & (cf->cond | cf->next))) {
+    if  (!(SU2_XFILE & cf->flags) && !((SU2_NOCHCK | SU2_UFOUND) & (cf->cond | cf->next))) {
         su2_log(cf, 0);
         ifBLIN_QX0("Denied");
         ex = EX_NOPERM;
         goto out;
     }
+#ifndef BLIN_OSLinux
     if  (!!(ex = su2_auth(cf))) {
         ifBLIN_QX0("Not preauthorized");
         ex = EX_NOPERM;
         goto out;
     }
-    if  (  !!cf->uid
-        && (  (cf->uid != cf->frompw->pw_uid)
-           || (~(cf->next | cf->cond) & SU2_NOPASS)
-           )
-        && !!(ex = su2_inpass(cf))
-        ) {
+#endif
+    do {
+        if  (!cf->uid) {
+            ifBLIN_QX2("OK root");
+            break;
+        }
+        if  (  (cf->uid == cf->frompw->pw_uid)
+            && (  ((cf->next | cf->cond) & SU2_NOPASS)
+               || !((SU2_XFILE | SU2_XPASS) & ~cf->flags)
+            )  ) {
+            ifBLIN_QX2("OK NOPASS %u %u %02X %02X", cf->frompw->pw_uid, cf->uid, cf->cond, cf->next);
+            break;
+        }
+        if  (!(ex = su2_inpass(cf))) {
+            ifBLIN_QX2("OK PASS");
+            break;
+        }
         su2_log(cf, 0);
         ifBLIN_QX0("Not authorized");
         ex = EX_NOPERM;
         goto out;
-    }
+    } while(0);
     su2_log(cf, 1);
     if  (!!(ex = su2_reg(cf, 1))) {
         ifBLIN_QW0("su2_reg in");
         goto out;
     }
+#ifndef BLIN_OSLinux
     if  (!cf->class) {
         if  (!(lc = login_getpwclass(cf->topw))) {
-            ifBLIN_QW0("login_getpwclass");
+            ifBLIN_QW0("login_getpwclass %u", cf->topw->pw_uid);
             ex = EX_OSERR;
             goto out;
         }
-    } else if (!cf->mypw->pw_uid && !cf->topw->pw_uid) {
-        ifBLIN_QX0("Only root may use -t");
+    } else if (!!cf->mypw->pw_uid || !!cf->topw->pw_uid) {
+        ifBLIN_QX0("Only root may use -t, now %u -> %u", cf->mypw->pw_uid, cf->topw->pw_uid);
         ex = EX_NOPERM;
         goto out;
     } else if (!(lc = login_getclass(cf->class))) {
         ifBLIN_QW0("login_getclass %s", cf->class);
         ex = EX_OSERR;
         goto out;
+    } else {
+        ifBLIN_QX6("login_getclass %s OK", cf->class);
     }
     if  (!!cf->class && (!lc->lc_class || !!strcmp(cf->class, lc->lc_class))) {
         ifBLIN_QX0("Unknown class: %s", cf->class);
@@ -436,6 +606,7 @@ main(int argc, char *argv[]) {
         ex = EX_OSERR;
         goto out;
     }
+#endif
     sa.sa_flags = SA_RESTART;
     sa.sa_handler = SIG_IGN;
     sigemptyset(&sa.sa_mask);
@@ -448,14 +619,18 @@ main(int argc, char *argv[]) {
     if  (pipe(fds) == -1) {
         ifBLIN_QX0("pipe");
         ex = EX_IOERR;
+#ifndef BLIN_OSLinux
         if  (!!su2_pamend(cf)) ifBLIN_QX1("su2_pamend");
+#endif
         goto out;
     }
     switch ((child = fork())) {
     case -1:
         ifBLIN_QW0("fork");
         ex = EX_OSERR;
+#ifndef BLIN_OSLinux
         if  (!!su2_pamend(cf)) ifBLIN_QX1("su2_pamend");
+#endif
         goto out;
     case 0 :
         close(fds[1]);
@@ -464,64 +639,35 @@ main(int argc, char *argv[]) {
         sigaction(SIGPIPE, &sa_pipe, NULL);
         sigaction(SIGINT, &sa_int, NULL);
         sigaction(SIGQUIT, &sa_quit, NULL);
-        if  (0 > setusercontext( lc
-                               , cf->topw
-                               , cf->topw->pw_uid
-                               , LOGIN_SETUSER | LOGIN_SETCPUMASK | LOGIN_SETLOGINCLASS
-            )                  ) {
-            ifBLIN_QW0("setusercontext(SETUSER SETCPUMASK SETLOGINCLASS)");
-            ex = EX_OSERR;
+        if  (!!(ex = su2_posenv( cf
+#ifndef BLIN_OSLinux
+                               , lc
+#endif
+            )  )               ) {
+            ifBLIN_QW0("su2_posenv");
             goto out;
         }
-        if  (SU2_LOGIN & cf->flags) {
-            const char    *p;
-            char         **e;
-            char          *v;
-
-            p = getenv("TERM");
-            environ = &cleanenv;
-            if  (!!setenv("HOME", cf->topw->pw_dir, 1)) {
-                ifBLIN_QW1("setenv(HOME, %s, 1)", cf->topw->pw_dir);
-            }
-            if  (!!setenv("SHELL", cf->shell, 1)) ifBLIN_QW1("setenv(SHELL, %s, 1)", cf->shell);
-            if  (!!(envpam = pam_getenvlist(cf->pamh))) {
-                for (e = envpam; !!*e; ++e) {
-                    if  (1024 < strlen(*e)) continue;
-                    if  (!strncmp(*e, "LD_", 3)) continue;
-                    if  (!(v = strchr(*e, '='))) continue;
-                    *(v++) = '\0';
-                    if  (1 == babolo_testword(&env, *e)) continue;
-                    if  (!!setenv(*e, v, 1)) ifBLIN_QW1("setenv(setenv(%s, %s, 1)", *e, v);
-            }   }
-            if  (0 > setusercontext( lc
-                                   , cf->topw
-                                   , cf->topw->pw_uid
-                                   , LOGIN_SETPATH | LOGIN_SETUMASK | LOGIN_SETENV | LOGIN_SETRESOURCES
-                )                  ) {
-                ifBLIN_QW0("setusercontext(LOGIN_SETPATH LOGIN_SETUMASK LOGIN_SETENV)");
-                ex = EX_OSERR;
-                goto out;
-            }
-            if  (!!p && !!setenv("TERM", p, 1)) ifBLIN_QW1("setenv(TERM, %s, 1)", p);
-            p = pam_getenv(cf->pamh, "HOME");
-            if  (!!chdir(p ? p : cf->topw->pw_dir)) {
-                ifBLIN_QW0("chdir %s", p ? p : cf->topw->pw_dir);
-                ex = EX_UNAVAILABLE;
-                goto out;
-        }   }
+#ifndef BLIN_OSLinux
         login_close(lc);
-        ifBLIN_QX3( "to execv %s(%s, %s, %s, %s)"
-                  , (SU2_DASH & cf->flags) ? cf->args[cf->argp] : cf->shell
-                  , ((cf->argp + 1) < cf->amax) ? cf->args[cf->argp + 1] : ""
-                  , ((cf->argp + 2) < cf->amax) ? cf->args[cf->argp + 2] : ""
-                  , ((cf->argp + 3) < cf->amax) ? cf->args[cf->argp + 3] : ""
-                  , ((cf->argp + 4) < cf->amax) ? cf->args[cf->argp + 4] : ""
-                  );
-        execv((SU2_DASH & cf->flags) ? cf->args[cf->argp] : cf->shell, &cf->args[cf->argp]);
-        ifBLIN_QX0("exec");
+#endif
+        if  ((SU2_DASH & cf->flags) || !!cf->xfile) cf->shell = cf->args[cf->argp];
+        ifBLIN_QX3("to execv %s(%d..%d)", !cf->shell ? "" : cf->shell, cf->argp, cf->amax);
+        for (uint i = 0; i < cf->amax; ++i) {
+            ifBLIN_QX3( "[%d]%c%s%s"
+                      , i
+                      , (i < cf->argp) ? '#' : '='
+                      , !cf->args[i] ? "" : cf->args[i]
+                      , !cf->args[i] ? "" : "~"
+                      );
+        }
+        execv(cf->shell, (char * const*)&cf->args[cf->argp]);
+        ifBLIN_QW0("exec=%s~", cf->shell);
         ex = EX_OSERR;
         goto out;
     default:
+#ifndef BLIN_OSLinux
+        login_close(lc);
+#endif
         sa.sa_handler = SIG_IGN;
         sigaction(SIGTTOU, &sa, NULL);
         close(fds[0]);
@@ -548,13 +694,17 @@ main(int argc, char *argv[]) {
         }
         if  (!!cf->tty) tcsetpgrp(cf->ttyfd, getpgrp());
         if  (pid == -1) err(1, "waitpid");
+#ifndef BLIN_OSLinux
         if  (!!su2_pamend(cf)) ifBLIN_QX1("su2_pamend");
+#endif
         if  (!!(ex = su2_reg(cf, 0))) {
             ifBLIN_QW0("su2_reg out");
             goto out;
         }
         ex = WEXITSTATUS(statusp);
     }
+    if  (!!cf->md && !!mife_fini(cf->md)) ifBLIN_QW0("mife_fini");
+    endpwent();
 out:
     exit(ex);
 #   undef blin_internal_flags
